/* SPDX-License-Identifier: BSD-3-Clause
 *
 * Copyright 2023 The ChromiumOS Authors
 * Use of this source code is governed by a BSD-style license that can be
 * found in the LICENSE file.
 */

 package gvip_axis_pkg;

  import gvip_policy_pkg::policy;
  import gvip_axis_item_pkg::generic_item;
  import gvip_axis_item_pkg::bitem;
  import gvip_axis_item_pkg::item;
  import greporter_pkg::greporter;
  import gvip_handshake_pkg::*;


  //////////////////////////////
  // Analysis ports interface //
  //////////////////////////////

  // Interface Class: generic_analysis_port_ic
  //
  // If Testbench operates generic_item, subscriber should implement this interface
  // Object with this interface can be register as Anaysis Port in VIP Core
  // Depending of registretion type write method will be called after each transfer
  // (core::add_generic_ap2transfer) or after whole packet (core::add_generic_ap2pkt)
  //
  interface class generic_analysis_port_ic;
    pure virtual function void write (input generic_item item);
  endclass : generic_analysis_port_ic


  // Interface Class: analysis_port_ic
  //
  // If Testbench operates item, subscriber should implement this interface
  // Object with this interface can be register as Anaysis Port in VIP Core
  // Depending of registretion type write method will be called after each transfer
  // (core::add_ap2transfer) or after whole packet (core::add_ap2pkt)
  //
  interface class analysis_port_ic #(type item_t = logic);
    pure virtual function void write (input item_t item);
  endclass : analysis_port_ic


  /////////////////////////////
  // VIP configuration types //
  /////////////////////////////

  // Enum: gvip_type_e
  //
  // This is used as VIP type choose
  // MASTER  - VIP will assert tvalid/tdata/tuser/tstrb/tkeep/tlast signals
  // SLAVE   - VIP will assert tready signal
  // MONITOR - VIP only observe bus and report transactions with analysis port
  //
  typedef enum logic [1:0] {
    MASTER,
    SLAVE,
    MONITOR
  } gvip_type_e;


  // Class: gvip_config
  //
  // Configuration object to configure VIP - can be pass to constructor
  //
  class gvip_config;

    // Variable: mask_data_with_keep_in_ap
    //
    // If this is set to 1, bytes in tdata with corresponding tkeep equal 0 will be cleared
    //
    bit mask_data_with_keep_in_ap = 0;

    // Variable: warn_resize_mismatch_from_generic
    //
    // If set to 0, no warning will be printed during conversion generic_item_t -> item_t
    // in case of size(generic_item_t::tdata) % N != 0
    //
    bit warn_resize_mismatch_from_generic = 1;

    // Variable: generic_item_squize_by_tkeep
    //
    // If this is set to 1, bytes of tdata  in generic_item with corresponding tkeep equal 0
    // will be removed - item will be have only valid bytes in tdata and all tkeep set to 1
    //
    bit generic_item_squize_by_tkeep = 0;

    // Variable: stop_vip
    //
    // If this is set to 1, VIP will not resume main thread after 'clear'
    // all VIPs thread will be stopped
    //
    bit stop_vip = 0;

    // Variable: transaction_in_reset
    //
    // In normal operation VIP operates only if reset in gclk_if is in not active state
    // If this variable is set, VIP does not take reset state into account
    //
    bit transaction_in_reset = 0;

    // Variable: report_id
    //
    // This string will be used as ID of messages generated by VIP
    //
    string report_id = "";

    // Variable: coverage
    //
    // Enable functional coverage counters
    // This value is taken into account only at vip core constructor
    //
    bit coverage = 0;

  endclass : gvip_config



  /////////////////////////////
  // Internal implementation //
  /////////////////////////////
  `include "gvip_axis_coverage.svh"
  `include "gvip_axis_imp.svh"
  `include "gvip_axis_policy.svh"


  /////////
  // VIP //
  /////////


  // Class: core
  //
  // VIP core class
  // parameters must be consistent with interface instance parameters
  //
  // Parameters:
  //   N   - number of bytes of tdata
  //   I   - tid signal width
  //   D   - tdest signal width
  //   U_T - tuser type
  //   item_t - item type definition that matches interface -
  //            for some simulators is requied to have single point of type definition
  //            so in such case type can by pass to VIP
  class core #(int N = 1, int I = 1, int D = 1, type U_T = logic [0:0],
               type item_t = gvip_axis_item_pkg::item #(N, I, D, U_T))
               extends core_imp #(N, I, D, U_T, item_t) implements generic_ic;

    // Function: new
    //
    // Creates an instance of VIP core
    //
    // name     - instance name
    // vif      - axi stream interface handler
    // vclk     - clock interface handler
    // vip_type - one of VIPs type (MASTER/SLAVE/MONITOR)
    // vip_cfg  - optional cofiguration object (please check: gvip_config)
    //
    extern function new (string name, virtual gaxis_if #(.N(N), .I(I), .D(D), .U_T(U_T)) vif,
                         virtual gclk_if vclk, gvip_type_e vip_type, gvip_config vip_cfg = null);

    // Function: get_data_width
    //
    // Returns width of tdata in bits
    //
    extern virtual function int unsigned get_data_width();

    // Function: get_user_width
    //
    // Returns width of tuser in bits
    //
    extern virtual function int unsigned get_user_width();

    // Function: get_keep_width
    //
    // Returns width of tkeep in bits
    //
    extern virtual function int unsigned get_keep_width();

    // Function: get_vip_type
    //
    // Returns type of VIP
    //
    extern virtual function gvip_type_e get_vip_type();

    // Function: get_config
    //
    // Returns current configuration object
    //
    extern virtual function gvip_config get_config();

    // Function: add_ap2pkt
    //
    // Use this function to register object with analysis port interface implementation,
    // `write` function from this interface will be called with tlast occurrence
    // Any number of object can be registered but since item does not deliver clone functionality
    // the same object instance will be passed to all subscribers
    //
    extern virtual function void add_ap2pkt(analysis_port_ic #(item_t) ap);

    // Function: add_ap2transfer
    //
    // Use this function to register object with analysis port interface implementation,
    // `write` function from this interface will be called with each transfer occurrence
    // Any number of object can be registered but since item does not deliver clone functionality
    // the same object instance will be passed to all subscribers
    //
    extern virtual function void add_ap2transfer(analysis_port_ic #(item_t) ap);

    // Function: add_generic_ap2pkt
    //
    // Use this function to register object with generic analysis port interface implementation,
    // `write` function from this interface will be called with tlast occurrence
    // Any number of object can be registered but since item does not deliver clone functionality
    // the same object instance will be passed to all subscribers
    //
    extern virtual function void add_generic_ap2pkt(generic_analysis_port_ic ap);

    // Function: add_generic_ap2transfer
    //
    // Use this function to register object with generic analysis port interface implementation,
    // `write` function from this interface will be called with each transfer occurrence
    // Any number of object can be registered but since item does not deliver clone functionality
    // the same object instance will be passed to all subscribers
    //
    extern virtual function void add_generic_ap2transfer(generic_analysis_port_ic ap);

    // Function: set_policy
    //
    // Assign new handshake policy to VIP
    //
    // obj - policy object derived from gvip_policy_pkg::policy
    // do_randomize_policy - if set, VIP will call randomize method at policy object,
    //                       usefull when policy is created without declaration e.g.:
    //                       vip.set_policy(gvip_policy_pkg::random::create("def_random"), 1);
    // clone - in default VIP will clone policy to operate on it
    //
    extern virtual function void set_policy(policy obj, bit do_randomize_policy = 0, bit clone = 1);

    // Function: item2generic
    //
    // Converts concrete item of this VIP to generic one
    //
    extern virtual function generic_item item2generic(item_t arg);

    // Function: generic2item
    //
    // Converts generic item to concrete item of this VIP
    //
    extern virtual function item_t generic2item(generic_item arg);

    // Function: send_item
    //
    // This task sends data item to AxiStream interface
    // If some package is currently trasmited, new item will be added to queue
    // depending ~item.is_posted~ field task ends immediately or block
    // Supports any type derived from bitem (e.g.  ~generic_item~ or ~item_t~)
    //
    extern virtual task send_item(bitem item);

    // Function: run_phase
    //
    // Main VIPs service loop - run VIPs processing (must be called by test)
    // This task spawn internal processes and return immediately
    // Spawned processes are endless unless ~gvip_config::stop_vip~ is set
    //
    extern virtual task run_phase();

    // Function: clean
    //
    // This function clean VIP
    // Breakes any transfers at interface, clear all queues and any data structures
    //
    extern virtual function void clean();

  endclass : core



  //////////////////////////////////
  // Core function implementation //
  //////////////////////////////////

  function core::new (string name,
                      virtual gaxis_if #(.N(N), .I(I), .D(D), .U_T(U_T)) vif,
                      virtual gclk_if vclk, gvip_type_e vip_type, gvip_config vip_cfg = null);
    this.name = name;
    this.vif = vif;
    this.vclk = vclk;
    this.vip_type = vip_type;
    this.cfg = vip_cfg;

    // user did not pass configuration, create default one
    if (cfg == null) cfg = new();

    // there is no report_id in configuration, create default one
    if (cfg.report_id.len() == 0) cfg.report_id = $sformatf("AXIS-VIP[%s]", name);

    // create monitor handshake object
    mon_hs = new(vif, vclk, cfg);

    // if configuratin request coverage, create it
    if (cfg.coverage) begin
      coverage = new(vif, vclk, name);
    end

    // depending on VIP type create master or slave handshake object
    if (vip_type == MASTER) begin
      hs = axis_master_hs #(N, I, D, U_T, item_t)::create(vif, vclk, cfg);
    end else if (vip_type == SLAVE) begin
      hs = axis_slave_hs #(N, I, D, U_T, item_t)::create(vif, vclk, cfg);
    end

    greporter::info(cfg.report_id, $sformatf("VIP '%s' CREATED", name));
  endfunction : new


  function generic_item core::item2generic(item_t arg);
    return gvip_axis_item_pkg::item_conv #(item_t)::item2generic(arg,
                                                    cfg.generic_item_squize_by_tkeep);
  endfunction : item2generic


  function core::item_t core::generic2item(generic_item arg);
    return gvip_axis_item_pkg::item_conv #(item_t)::generic2item(arg,
                                                    cfg.warn_resize_mismatch_from_generic,
                                                    cfg.generic_item_squize_by_tkeep);
  endfunction : generic2item


  function void core::add_ap2pkt(analysis_port_ic #(item_t) ap);
    mon_hs.add_ap2pkt(ap);
  endfunction : add_ap2pkt


  function void core::add_ap2transfer(analysis_port_ic #(item_t) ap);
    mon_hs.add_ap2transfer(ap);
  endfunction : add_ap2transfer


  function void core::add_generic_ap2pkt(generic_analysis_port_ic ap);
    mon_hs.add_ap2pkt(to_generic_item_ap #(item_t)::create(ap, cfg));
  endfunction : add_generic_ap2pkt


  function void core::add_generic_ap2transfer(generic_analysis_port_ic ap);
    mon_hs.add_ap2transfer(to_generic_item_ap #(item_t)::create(ap, cfg));
  endfunction : add_generic_ap2transfer


  function int unsigned core::get_data_width();
    return (vif.N * 8);
  endfunction : get_data_width


  function int unsigned core::get_user_width();
    return vif.U_T_WIDTH;
  endfunction : get_user_width


  function int unsigned core::get_keep_width();
    return vif.N;
  endfunction : get_keep_width


  function gvip_type_e core::get_vip_type();
    return vip_type;
  endfunction : get_vip_type


  function void core::set_policy(policy obj, bit do_randomize_policy = 0, bit clone = 1);
    if (do_randomize_policy && !obj.randomize()) begin
      greporter::error(cfg.report_id, $sformatf("Policy '%s' randomizastion issue", obj.name));
    end
    greporter::info(cfg.report_id, $sformatf("Policy set to %s [%s]", obj.get_name(),
                                                                      obj.get_type_name()));
    hs.set_policy(obj, clone);
  endfunction : set_policy


  task core::run_phase();
    // enable interface access
    case (vip_type)
      MASTER : vif.g_sim.force_as_master();
      SLAVE  : vif.g_sim.force_as_slave();
      default : begin end
    endcase

    // start core threads
    fork
      if (hs != null) begin
        while(!cfg.stop_vip) hs.start(cfg.transaction_in_reset);
      end

      while(!cfg.stop_vip) mon_hs.start(cfg.transaction_in_reset);
    join_none
  endtask : run_phase


  task core::send_item(bitem item);
    axis_master_hs #(N, I, D, U_T) m;

    if (vip_type != MASTER) begin // only master supports data send
      greporter::fatal(cfg.report_id, "This is not master VIP");
    end else if (!$cast(m, hs)) begin
      // VIP type is master but for some reason we can't cast
      // this should never happen
      greporter::fatal(cfg.report_id, "Wrong master VIP type");
    end

    // put requested item to queue, main thread from hs will take this item and put on bus
    m.put(item);

    // if item is not posted, block this task till item packet will be putted on bus
    if (!item.is_posted) begin
      @item.post;
    end
  endtask : send_item


  function gvip_config core::get_config();
    return cfg;
  endfunction : get_config


  function void core::clean();
    hs.clean();
    mon_hs.clean();
  endfunction : clean


endpackage : gvip_axis_pkg
